#ifndef PODOFO_CONFIG_H
#define PODOFO_CONFIG_H

// Template filled out by CMake

#define PODOFO_VERSION_MAJOR 1
#define PODOFO_VERSION_MINOR 1
#define PODOFO_VERSION_PATCH 0

// PoDoFo configuration options

// Libraries
#define PODOFO_HAVE_JPEG_LIB
#define PODOFO_HAVE_PNG_LIB
#define PODOFO_HAVE_TIFF_LIB
#define PODOFO_HAVE_FONTCONFIG
/* #undef PODOFO_HAVE_WIN32GDI */

#endif // PODOFO_CONFIG_H
